from io import BytesIO
import re
import sys
from typing import List

import networkx as nx

from peach.constants import Mode
from peach.shared_state import shared_state
from peach.table import Table


def _build_dag(tables: List[Table]) -> nx.DiGraph:
    dag = nx.DiGraph()
    dag.add_nodes_from(d.id for d in tables)
    dag.add_edges_from((i, d.id) for d in tables for i in d.inputs)
    return dag


MODE_TO_COLOR_MAP = {
    Mode.PROD: "#ffffff",
    Mode.DEV: "#51c991",
}


def give_up_and_exit(message: str, exit_code: int = 1) -> None:
    print(message, file=sys.stderr)
    sys.exit(exit_code)


def _draw_dag(dag: nx.DiGraph, title: str = None, full_dag: nx.DiGraph = None) -> None:
    """

    Requires pygraphviz+graphviz, and either Pillow or matplotlib to be installed.
    """
    try:
        agraph = nx.nx_agraph.to_agraph(full_dag or dag)
    except ImportError:
        give_up_and_exit(
            "In order to draw DAGs you must install graphviz. Follow the instructions here:"
        )
        raise

    agraph.node_attr["shape"] = "box"
    agraph.node_attr["style"] = "filled"

    for node in agraph:
        table = shared_state.tables[node.name]
        label = "External"
        if not table.is_external:
            label = shared_state.jobs[node.name].__class__

        mode = table.mode
        if table.is_external:
            mode = Mode.PROD
        node.attr["fillcolor"] = MODE_TO_COLOR_MAP[mode]

        if full_dag and node not in dag:
            node.attr["style"] = "dashed"

        job_class = (
            str(label)
            .removeprefix("<class 'peach.jobs.")
            .removesuffix("'>")
        )
        node.attr["label"] = f"{node.name}\n({job_class})"

    # add an ugly legend
    agraph.add_node("Not selected, Mode.PROD", style="dashed")
    agraph.add_node(
        "Selected, Mode.PROD", fillcolor=MODE_TO_COLOR_MAP[Mode.PROD], style="filled"
    )
    agraph.add_node(
        "Selected, Mode.DEV", fillcolor=MODE_TO_COLOR_MAP[Mode.DEV], style="filled"
    )
    agraph.add_edge("Not selected, Mode.PROD", "Selected, Mode.PROD", style="invis")
    agraph.add_edge("Selected, Mode.PROD", "Selected, Mode.DEV", style="invis")
    agraph.add_subgraph(
        ["Not selected, Mode.PROD", "Selected, Mode.PROD", "Selected, Mode.DEV"],
        fillcolor="#bbbbbb",
        label="Legend",
        name="cluster_legend",
        shape="rect",
        style="filled",
    )

    agraph.layout("dot")
    img = agraph.draw(format="png")

    try:
        from PIL import Image

        image = Image.open(BytesIO(img))
        image.show(title=title or "DAG")
    except ModuleNotFoundError:
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg

        plt.figure()
        plt.imshow(mpimg.imread(BytesIO(img), format="png"))
        plt.show()


def dag() -> nx.DiGraph:
    return _build_dag(shared_state.tables.values())


def dag_to_steps(dag: nx.DiGraph) -> List[List[str]]:
    return list(nx.topological_generations(dag))


def steps_to_subdag(super_dag: nx.DiGraph, steps: List[List[str]]) -> nx.DiGraph:
    flat_steps = [ds for sublist in steps for ds in sublist]
    return nx.subgraph(super_dag, flat_steps)


def ancestor_steps(
    dag: nx.DiGraph, table_id: str, limit: int = None
) -> List[List[str]]:
    steps = dag_to_steps(nx.subgraph(dag, nx.ancestors(dag, table_id)))
    if limit is None:
        return steps
    else:
        return steps[-limit:]


def descendant_steps(
    dag: nx.DiGraph, table_id: str, limit: int = None
) -> List[List[str]]:
    steps = dag_to_steps(nx.subgraph(dag, nx.descendants(dag, table_id)))
    if limit is None:
        return steps
    else:
        return steps[:limit]


def retrieve_subdag(dag: nx.DiGraph, query_string: str) -> nx.DiGraph:
    # example query_strings and what they would match:
    # a/table/id     - match exactly that single node in the DAG
    # +a/table/id    - match that node, and everything upstream from it
    # 2+a/table/id   - match that node, and everything upstream from it max 2 steps removed
    # a/table/id+2   - same, but downstream
    # 2+a/table/id+2 - same, but both upstream and downstream
    anc, anc_count, table_id, des, des_count = re.match(
        r"(([0-9]*)\+)?([a-zA-Z/_]+)(\+([0-9]*))?", query_string
    ).groups()

    anc_steps = []
    des_steps = []
    if anc:
        anc_count = int(anc_count) if anc_count != "" else None
        anc_steps = ancestor_steps(dag, table_id, anc_count)
    if des:
        des_count = int(des_count) if des_count != "" else None
        des_steps = descendant_steps(dag, table_id, des_count)

    return steps_to_subdag(dag, anc_steps + [[table_id]] + des_steps)
