from dataclasses import dataclass, field
from functools import cached_property
import os
from pathlib import Path
import tempfile
from typing import Dict, Tuple

import google.api_core.exceptions
from google.cloud import bigquery

import pyarrow as pa
from pyarrow import csv as pa_csv
from pyarrow import parquet as pa_pq

from peach.constants import Mode, UpdateMethod
from peach.exceptions import CodeException, IncompatibleSchemaException
from peach.table import Table


@dataclass(frozen=True)
class Datastore:
    id: str

    def locator_string(self, table: Table) -> str:
        raise NotImplementedError()

    def read_as_arrow(self, table: Table) -> pa.Table:
        raise NotImplementedError()

    def write_from_arrow(self, table: Table, data: pa.Table, update_method: UpdateMethod) -> None:
        raise NotImplementedError()

    def write_from_sql_query(self, table: Table, sql_query: str, update_method: UpdateMethod) -> None:
        raise NotImplementedError()


@dataclass(frozen=True)
class DirectoryDatastore(Datastore):
    directory: str = "directory_datastore"

    def locator_string(self, table: Table) -> str:
        return str(Path(self.directory, Path(table.id).with_suffix(".csv")).resolve())

    def read_as_arrow(self, table: Table) -> pa.Table:
        file_path = Path(self.locator_string(table))
        data = pa_csv.read_csv(file_path)
        data = data.cast(table.schema)
        return data

    def write_from_arrow(self, table: Table, data: pa.Table, update_method: UpdateMethod) -> None:
        data = data.cast(table.schema)

        file_path = Path(self.locator_string(table))
        file_path.parent.mkdir(parents=True, exist_ok=True)

        mode = "w" if update_method == UpdateMethod.REPLACE else "a"
        include_header = update_method == UpdateMethod.REPLACE
        options = pa_csv.WriteOptions(include_header=include_header)
        with open(file_path, mode) as f:
            pa_csv.write_csv(data=data, output_file=f, write_options=options)


@dataclass(frozen=True)
class BigQueryDatastore(Datastore):
    default_project_id: str
    default_location: str
    default_compute_region: str
    file_temp_location: str
    dev_id: str

    @cached_property
    def _bqclient(self) -> bigquery.Client:
        return bigquery.Client()

    def _create_schema(
        self,
        schema_name: str,
        exists_ok: bool = True,
    ) -> None:
        dataset = bigquery.Dataset(schema_name)
        dataset.location = self.default_location
        self._bqclient.create_dataset(dataset, exists_ok=exists_ok, timeout=30)

    def locator_string(self, table: Table) -> str:
        """Map table to BigQuery table reference.

        Details:
            - Underscores in the first folder are converted to hyphens for the following reasons:
              1) Google Cloud project can contain hyphens, but not underscores.
              2) For Python source imports it's bad to use hyphens while imporing packages.

        Example:
            In prod turns "a_A/b/c/d" into "a-A.b.c__d".
            In dev turns "a_A/b/c/d" into "a-A.erik.b__c__d".
        """
        if table.mode is Mode.PROD:
            parts = table.id.split("/")
            if len(parts) < 3:
                raise ValueError(f"Table id should have at least 3 components in prod mode: {table.id}")

            project_id = parts[0].replace("_", "-")
            dataset_id = parts[1]
            table_id = "__".join(parts[2:])
            return f"{project_id}.{dataset_id}.{table_id}"

        if table.mode is Mode.DEV:
            parts = table.id.split("/")
            if len(parts) < 2:
                raise ValueError(f"Table id should have at least 2 components in dev mode: {table.id}")

            project_id = parts[0].replace("_", "-")
            dataset_id = self.dev_id
            table_id = "__".join(parts[1:])
            return f"{project_id}.{dataset_id}.{table_id}"

        raise NotImplementedError(f"Unsupported table mode: {table.mode}")

    def locator_parts(self, locator_string: str) -> Tuple[str, str, str]:
        parts = locator_string.split(".")
        if len(parts) != 3:
            raise ValueError(f"Locator string should have exactly 3 parts: {locator_string}")

        project, schema, table = parts[0], parts[1], parts[2]
        return project, schema, table

    def read_as_arrow(self, table: Table) -> pa.Table:
        locator_string = self.locator_string(table)
        sql = f"select * from {locator_string}"
        query_job = self._bqclient.query(sql)
        data = query_job.to_arrow()

        if not data.schema.equals(table.schema):
            data_schema_extra = set(data.schema) - set(table.schema)
            def_schema_extra = set(table.schema) - set(data.schema)
            raise IncompatibleSchemaException(
                f"Table schema fields: {def_schema_extra} differ from "
                f"BigQuery schema: {data_schema_extra}"
            )

        return data

    def write_from_arrow(self, table: Table, data: pa.Table, update_method: UpdateMethod) -> None:
        if not data.schema.equals(table.schema):
            data_schema_extra = set(data.schema) - set(table.schema)
            def_schema_extra = set(table.schema) - set(data.schema)
            raise IncompatibleSchemaException(
                f"Table schema fields: {def_schema_extra} differ from "
                f"Produced data schema: {data_schema_extra}"
            )

        table_ref = self.locator_string(table)
        project, schema, _ = self.locator_parts(table_ref)
        self._create_schema(f"{project}.{schema}")

        write_disposition = "WRITE_TRUNCATE"
        if update_method == UpdateMethod.APPEND:
            write_disposition = "WRITE_APPEND"

        parquet_options=bigquery.ParquetOptions()
        parquet_options.enable_list_inference = True
        load_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.PARQUET,
            parquet_options=parquet_options,
            create_disposition="CREATE_IF_NEEDED",
            write_disposition=write_disposition,
        )

        tmpfd, tmppath = tempfile.mkstemp(suffix=".parquet")
        os.close(tmpfd)

        try:
            pa_pq.write_table(data, tmppath)
            with open(tmppath, "rb") as tmpfile:
                file_size = os.path.getsize(tmppath)
                load_job = self._bqclient.load_table_from_file(
                    file_obj=tmpfile,
                    destination=table_ref,
                    size=file_size,
                    job_config=load_config,
                )
                load_job.result()
        except google.api_core.exceptions.BadRequest as ex:
            raise CodeException() from ex
        finally:
            os.remove(tmppath)

    def write_from_sql_query(self, table: Table, sql_query: str, update_method: UpdateMethod) -> None:
        table_ref = self.locator_string(table)

        project, schema, _ = self.locator_parts(table_ref)
        self._create_schema(f"{project}.{schema}")

        try:
            # Store query result to BQ temp table, so that we can do schema compatibility check
            # before updating output table.
            query_job = self._bqclient.query(sql_query)
            query_result = query_job.result()

            # Schema compatibility check.
            query_iterator = query_result.to_arrow_iterable()
            query_batch = next(query_iterator)
            if not table.schema.equals(query_batch.schema):
                sql_schema_extra = set(query_batch.schema) - set(table.schema)
                def_schema_extra = set(table.schema) - set(query_batch.schema)
                raise IncompatibleSchemaException(
                    f"Table schema fields: {def_schema_extra} differ from "
                    f"SQL query schema: {sql_schema_extra}"
                )

            # Copy temp table to final destination.
            write_disposition = "WRITE_TRUNCATE"
            if update_method == UpdateMethod.APPEND:
                write_disposition = "WRITE_APPEND"
            copy_job = self._bqclient.copy_table(
                sources=query_result._table,
                destination=table_ref,
                job_config=bigquery.CopyJobConfig(
                    create_disposition="CREATE_IF_NEEDED",
                    write_disposition=write_disposition,
                ),
            )
            copy_job.result()
        except google.api_core.exceptions.BadRequest as ex:
            raise CodeException() from ex
