import logging
import os
import importlib
import subprocess
import sys
import uuid
from pathlib import Path
from typing import Optional

import fire
import networkx as nx

import peach.dag
from peach.constants import Mode
from peach.exceptions import (
    CLIArgumentException,
    ConfigException,
    CommandException,
    ProjectStructureException,
)
from peach.shared_state import shared_state
from peach.table import TABLES_DIR


MANIFEST_FILENAME = "manifest.py"
PYTHON_JOB_FILENAME = "job.py"


def import_module_from_path(file_path: str) -> None:
    importlib.import_module(file_path)


def import_manifest(file_path: str, with_job: bool) -> None:
    if not os.path.exists(file_path):
        raise ProjectStructureException(f"Missing manifest file under: {file_path}")
    import_module_from_path(str(file_path)[:-3].replace("/", "."))

    if with_job:
        dir_path = os.path.dirname(file_path)
        job_path = os.path.join(dir_path, PYTHON_JOB_FILENAME)
        if os.path.exists(job_path):
            import_module_from_path(job_path[:-3].replace("/", "."))


def init_configs(project_path) -> None:
    os.chdir(project_path)
    sys.path.append(".")

    import_module_from_path("local_config")
    import_module_from_path("project_structure")


def crawl_manifest_with_deps(job_path: str) -> None:
    manifest_full_path = os.path.join(TABLES_DIR, job_path, MANIFEST_FILENAME)
    logging.info(f"Parsing manifest: {manifest_full_path}")
    import_manifest(manifest_full_path, with_job=True)
    logging.info(f"Parsing manifest {manifest_full_path} done!")

    if len(shared_state.tables) != 1:
        raise ProjectStructureException(
            f"Should declare exactly one table per manifest: {manifest_full_path}")
    table = next(iter(shared_state.tables.values()))

    logging.info("Parsing manifests for dependencies...")
    for input_tid in table.inputs:
        manifest_path = os.path.join(TABLES_DIR, input_tid, MANIFEST_FILENAME)
        import_manifest(manifest_path, with_job=False)
    logging.info("Parsing manifests for dependencies done!")


def crawl_manifests() -> None:
    logging.info("Parsing manifests...")
    manifest_files = Path(TABLES_DIR).glob(f"**/{MANIFEST_FILENAME}")
    for file_path in manifest_files:
        logging.info(f"Parsing manifest: {file_path}")
        import_manifest(file_path, with_job=True)
    logging.info("Parsing manifests done!")

    logging.info("Verifying dependencies...")
    for tid, table in shared_state.tables.items():
        for input_tid in table.inputs:
            if input_tid not in shared_state.tables.keys():
                logging.error(
                    f"Error! {tid} depends on {input_tid}, which isn't declared!"
                )
    logging.info("Verifying dependencies done!")


def get_full_dag() -> nx.DiGraph:
    dag = peach.dag.dag()
    if not nx.is_directed_acyclic_graph(dag):
        raise ProjectStructureException("Detected loops in project structure dependencies")
    return dag


def run_process(cmd):
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    while True:
        line = process.stdout.readline()
        if not line:
            break
        logging.info(line)


    return_code = process.poll()
    if return_code != 0:
        raise CommandException(f"Failed to execute: {cmd}, returncode: {return_code}")


def build_image(
    docker_repository: str,
    dockerfile_path: str,
    requirements_path: str,
):
    docker_image_tag = os.path.basename(os.getcwd())
    requirements_dir = os.path.dirname(requirements_path)
    if requirements_dir:
        docker_image_suffix = "__".join(requirements_dir.split("/"))
        docker_image_tag = f"{docker_image_tag}__{docker_image_suffix}"
    if docker_repository:
        docker_image_tag = f"{docker_repository}/{docker_image_tag}"

    logging.info(
        f"Building image: {docker_image_tag} "
        f"from Dockerfile: {dockerfile_path} "
        f"with requirements: {requirements_path}..."
    )

    docker_build_cmd = [
        "docker", "build", ".",
        "--file", dockerfile_path,
        "--tag", docker_image_tag,
        "--platform", "linux/amd64",
        "--build-arg", "ENV=PROD",
        "--build-arg", f"REQUIREMENTS_PATH={requirements_path}"
    ]

    run_process(docker_build_cmd)

    return docker_image_tag


def push_image(image_tag):
    logging.info(f"Pushing image: {image_tag}")
    docker_push_cmd = [
        "docker", "push", image_tag,
    ]
    run_process(docker_push_cmd)


def run_image(image_tag, cmd, args, cloud=False):
    logging.info(f"Running image: {image_tag} with cli: {cmd} {args}")

    if not cloud:
        docker_run_cmd = [
            "docker", "run", image_tag, cmd, *args
        ]
        run_process(docker_run_cmd)
        return

    job_uuid = str(uuid.uuid4())
    cmd_args = [f"--args={arg}" for arg in args]
    cloud_run_cmd = [
        "gcloud", "beta", "run", "jobs", "create",
        f"job-{job_uuid}",
        f"--command={cmd}",
        *cmd_args,
        f"--image={image_tag}",
        f"--region=europe-west9",
        f"--cpu=2.0",
        f"--memory=4Gi",
        f"--execute-now",
        f"--wait",
        f"--max-retries=0",
    ]
    run_process(cloud_run_cmd)


def get_selected_sub_dag(
        full_dag: nx.DiGraph, dag_query_string: str, mode: Mode
    ) -> nx.DiGraph:
        sub_dag = full_dag
        if dag_query_string:
            sub_dag = peach.dag.retrieve_subdag(full_dag, dag_query_string)

        for table_id in sub_dag:
            shared_state.tables[table_id].mode = mode

        return sub_dag


def force_find_file(path: str, filename: str) -> str:
    path = os.path.join(TABLES_DIR, path)
    while path:
        if os.path.exists(os.path.join(path, filename)):
            break
        path = os.path.dirname(path)
    path = os.path.join(path, filename)
    if not os.path.exists(path):
        raise ProjectStructureException(
            f"Project should contain at least one {filename} under path {path} or its' parents. "
            f"Found none"
        )
    return path


class CliCommands:

    @staticmethod
    def run(
        project_path: str,
        dag_query_string: Optional[str] = None,
        mode: str = "dev",
        cloud: bool = False,
    ) -> None:
        if not Mode.is_valid(mode.upper()):
            raise CLIArgumentException(
                f"Invalid mode='{mode}' (valid modes={Mode.values()}")
        shared_state.mode = Mode(mode.upper())

        init_configs(project_path)
        crawl_manifests()

        dag = get_full_dag()
        dag = get_selected_sub_dag(dag, dag_query_string, Mode(mode.upper()))

        logging.info("Starting to run jobs...")
        steps = peach.dag.dag_to_steps(dag)

        for step in steps:
            for table_id in step:
                if table_id not in shared_state.jobs:
                    logging.info(f"External artifact exists for {table_id}")
                    continue
                job = shared_state.jobs[table_id]

                logging.info(f"Building artifact for {table_id}...")
                dockerfile_path = force_find_file(table_id, "Dockerfile")
                requirements_path = force_find_file(table_id, "requirements.txt")
                img_tag = build_image(
                    docker_repository=shared_state.docker_repository,
                    dockerfile_path=dockerfile_path,
                    requirements_path=requirements_path,
                )
                cmd = "python"
                args = [
                    "-m", "peach.pli", "run_job",
                    f"--project_path=.",
                    f"--job_path={table_id}",
                    f"--mode={mode}",
                    f"--containerized=False",
                    f"--cloud=False",
                ]
                if job.is_cli():
                    args = [
                        "-m", os.path.join(table_id, PYTHON_JOB_FILENAME),
                        *job.args,
                    ]
                if cloud:
                    if not shared_state.docker_repository:
                        raise ConfigException("No docker_repository defined in config")
                    push_image(img_tag)
                run_image(img_tag, cmd=cmd, args=args, cloud=cloud)
                logging.info(f"Done building artifact for {table_id}!")

        logging.info("Finished running all jobs!")


    @staticmethod
    def run_job(
        project_path: str,
        job_path: str,
        mode: str = "dev",
        containerized: bool = True,
        cloud: bool = False,
    ) -> None:
        if not Mode.is_valid(mode.upper()):
            raise CLIArgumentException(
                f"Invalid mode='{mode}' (valid modes={Mode.values()}")
        shared_state.mode = Mode(mode.upper())

        init_configs(project_path)
        crawl_manifest_with_deps(job_path)

        for table_id in shared_state.tables.keys():
            table = shared_state.tables[table_id]
            if table.is_external:
                table.mode = Mode.PROD
            else:
                table.mode = Mode(mode.upper())

        if len(shared_state.jobs) != 1:
            raise ProjectStructureException(
                f"Should declare exactly one job per path: {job_path}")
        table_id, job = next(iter(shared_state.jobs.items()))

        if containerized or cloud:
            logging.info("Running...")
            dockerfile_path = force_find_file(job_path, "Dockerfile")
            requirements_path = force_find_file(job_path, "requirements.txt")
            img_tag = build_image(
                docker_repository=shared_state.docker_repository,
                dockerfile_path=dockerfile_path,
                requirements_path=requirements_path,
            )

            if cloud:
                if not shared_state.docker_repository:
                    raise CLIArgumentException(
                        "Argument docker_repository is required with cloud option")
                push_image(img_tag)

            cmd = "python"
            args = [
                "-m", "peach.pli", "run_job",
                f"--project_path=.",
                f"--job_path={job_path}",
                f"--mode={mode}",
                f"--containerized=False",
                f"--cloud=False",
            ]
            if job.is_cli():
                args = [
                    "-m", os.path.join(job_path, PYTHON_JOB_FILENAME),
                    *job.args,
                ]
            run_image(img_tag, cmd=cmd, args=args, cloud=cloud)
            return

        if job.is_cli():
            raise CommandException("Cannot run CLI job without container")

        datastore = shared_state.datastore
        table = shared_state.tables[table_id]
        input_tables = shared_state.input_tables(table)

        logging.info(f"Updating table '{table.id}' (mode={table.mode})...")
        job.update_table(datastore, table, input_tables)
        logging.info(f"Updating table '{table.id}' (mode={table.mode}) done!")


    @staticmethod
    def draw_dag(
        project_path: str,
        dag_query_string: Optional[str] = None,
        mode: str = "dev",
        full: bool = False,
    ) -> None:
        if not Mode.is_valid(mode.upper()):
            raise CLIArgumentException(
                f"Invalid mode='{mode}' (valid modes={Mode.values()}")
        shared_state.mode = Mode(mode.upper())

        init_configs(project_path)
        crawl_manifests()

        full_dag = get_full_dag()
        selected_sub_dag = get_selected_sub_dag(
            full_dag, dag_query_string, Mode(mode.upper())
        )

        if full:
            peach.dag._draw_dag(selected_sub_dag, full_dag=full_dag)
        else:
            peach.dag._draw_dag(selected_sub_dag, full_dag=None)


def main():
    logging.getLogger().setLevel(logging.INFO)
    logging.info("Starting to peach!")
    fire.Fire(CliCommands)


if __name__ == "__main__":
    main()
